#include "conversions.h"
#include "myExcepts.h"

#include <iostream>
#include <fstream>
#include <cmath>


/**
 * construct map filename from input string
 * @param inputfn
 * @param ext
 * @return 
 */
std::string mapfilename(const std::string& inputfn, const std::string& ext) {
    std::string inp (inputfn);
    size_t lastslash = inp.rfind('/');
    size_t lastdot   = inp.rfind('.');
    
    if (lastdot == std::string::npos) lastdot = inp.length();
    
    if (lastslash != std::string::npos) {
        inp = inp.substr(lastslash+1, lastdot - lastslash -1);
    }
    else {
        inp = inp.substr(0, lastdot - lastslash -1);
    }
    inp += ext;
    
    return inp;
}

/**
 * create clipper Spacegroup from list of symops.
 * @param symopsStrings
 * @return 
 */
clipper::Spacegroup clipperSG(const std::vector<std::string>& symopsStrings) {
    std::string symopsstring (symopsStrings.front());
    
    std::vector<std::string>::const_iterator it;
    for (it=symopsStrings.begin()+1; it != symopsStrings.end(); ++it) {
        symopsstring += ";";
        symopsstring += *it;
    }
    clipper::Spgr_descr sgdescr(symopsstring);
    
    return clipper::Spacegroup (sgdescr);
}

/**
 * Convert parameters to clipper Cell
 * @param a
 * @param b
 * @param c
 * @param alpha
 * @param beta
 * @param gamma
 * @return 
 */
clipper::Cell clipperCell(float a, float b, float c, 
        float alpha, float beta, float gamma) {
    
    clipper::Cell_descr cdescr (a,b,c, alpha, beta, gamma);
    return clipper::Cell (cdescr);    
}


void syminfoFromPDB(const std::string& pdbfile, 
        clipper::Spacegroup& sg, 
        clipper::Cell& cell) {
    
    std::string cryst1card;
    char d(0);
    std::string sgsymbol;
    double a, b, c, alpha, beta, gamma;
    std::ifstream inp (pdbfile.c_str());
    if (! inp.is_open()) {
        std::cout << "*** Error: unable to open PDB-file " << pdbfile << "\n";
        throw myExcepts::IO::Filename (pdbfile.c_str());
    }
    
    while (cryst1card != "CRYST1") {
        inp >> cryst1card;
        if (inp.fail()) {
            std::cout << "*** Error: unable to find CRYST1 card in PDB-file"
                    << pdbfile << "\n";
            break;
        }
    }
    
    inp >> a >> b >> c >> alpha >> beta >> gamma;
    
    if (inp.fail()) {
        std::cout << "**** Error extracting unit cell parameters from "
                << "CRYST1 card.\n";
        throw myExcepts::IO::Filename (pdbfile.c_str());
    }
    cell = clipperCell(a, b, c, alpha, beta, gamma);
    
    while (d != '\n') {
        inp >> std::noskipws >> d;
        if (inp.fail() || d == '\n') break;
        sgsymbol.push_back(d);
    }
    if (inp.fail()) {
        std::cout << "**** Error extracting unit cell parameters from "
                << "CRYST1 card.\n";
        throw myExcepts::IO::Filename (pdbfile.c_str());
    }
    clipper::Spgr_descr sgdescr(sgsymbol);
    sg = clipper::Spacegroup(sgdescr);
    
    inp.close();
}

/**
 * 
 * @param a length of a
 * @param b length of b
 * @param c length of c
 * @param alpha angle in rad
 * @param beta  angle in rad
 * @param gamma angle in rad
 * @param avec 
 * @param bvec
 * @param cvec
 */
void cellVectors(const double a, const double b, const double c, 
        const double alpha, const double beta, const double gamma, 
        Vec3& avec, Vec3& bvec, Vec3& cvec) {
    avec = Vec3(a, 0,0);
    
    const double bx (b*std::cos(gamma));
    const double by (b*std::sin(gamma));
    bvec = Vec3(bx, by, 0.0);
    
    const double cx (c*std::cos(beta));
    const double cy ( (b*c*std::cos(alpha) - bx*cx)/by);
    double cz = c*c - cx*cx - cy*cy;
    cz = std::sqrt(cz);
    
    cvec = Vec3(cx, cy, cz);
}

void recCell(const Vec3& a, const Vec3& b, const Vec3& c, 
        Vec3& astar, Vec3& bstar, Vec3& cstar) {
    double Vstar (cross (a, b) * c);
    Vstar = 1.0/Vstar;
    Vstar = std::abs (Vstar);
    
    astar = Vstar* cross (b, c);
    bstar = Vstar* cross (c, a);
    cstar = Vstar* cross (a, b);
    
}