/* 
 * File:   fcffile.cpp
 * Author: tg
 * 
 * Created on July 9, 2012, 8:05 PM
 */

#include "fcffile.h"
#include "myExcepts.h"
#include "ReflUtils.h"

#include <fstream>
#include <iostream>
#include <sstream>
#include <cmath>
#include <algorithm>

FCFfile::FCFfile(const char* filename, 
        const double downweight)
: mapcoeffile(filename), weightWeaks_(downweight) {
    try {
        read_fcf(filename);
    }
    catch (...) {
        throw;
    }
}

FCFfile::~FCFfile() {
}

void FCFfile::read_fcf(const char* filename) {
    std::ifstream inp (filename);
    
    if (!inp.is_open()) {
        std::string fn (filename);
        fn += ".fcf";
        inp.open(fn.c_str());
        if (!inp.is_open()) {
            std::cerr << "*** Error: unable open "
                    << filename 
                    << " or " << fn << "\n";
        return;
        }
        /* update filename to what could have been opened */
        this->filename(fn);
    }

    read_fcfheader(inp);
    read_fcfdata(inp);
    
    inp.close();
}

void FCFfile::read_fcfheader(std::ifstream& inp) {
    const unsigned int linelength (81);
    char line[linelength];

    int listcode (-1);
    double Fcalcmax;
    double F000;
    /*
     *  some initialisations
     */
    dhighres_ = -1.0;
    while (true) {
        inp.getline(line, linelength);
        std::istringstream conv (line);
        std::string dummy;
        conv >> dummy;
        // last keyword: break and read data
        /* some short-cuts */
        if (dummy.empty() || dummy.at(0) == '#') continue;
        if (dummy == "_refln_phase_calc") break;
        if (dummy == "_shelx_refln_list_code") {
            conv >> listcode;
            if (listcode != 6) {
                std::ostringstream ostr;
                ostr << listcode;
                throw myExcepts::Format::FCF::ListCode(ostr.str().c_str());
            }
            else continue;
        }
        if (dummy == "_cell_length_a") {
            conv >> a_; 
            continue;
        }
        if (dummy == "_cell_length_b") {
            conv >> b_; 
            continue;
        }
        if (dummy == "_cell_length_c") {
            conv >> c_; 
            continue;
        }
        if (dummy == "_cell_angle_alpha") {
            conv >> alpha_; 
            continue;
        }
        if (dummy == "_cell_angle_beta") {
            conv >> beta_; 
            continue;
        }
        if (dummy == "_cell_angle_gamma") {
            conv >> gamma_; 
            continue;
        }
        if (dummy == "_shelx_F_calc_maximum") {
            conv >> Fcalcmax; 
            continue;
        }
        if (dummy == "_exptl_crystal_F_000") {
            conv >> F000; 
            continue;
        }
        if (dummy == "_reflns_d_resolution_high") {
            conv >> dhighres_; 
            continue;
        }
        if (dummy == "_symmetry_equiv_pos_as_xyz" ||
                dummy == "_space_group_symop_operation_xyz") {
            readsymop(inp);
            continue;
        }
    }
}

void FCFfile::read_fcfdata(std::ifstream& inp) {
    /* data read from LIST 6 - fcf file */
    int h,k,l;
    double Imeas; //! actually F^2
    double sigImeas;
    double Fcalc;
    float  phicalc;

    // finally read the data
    while (inp.good()) {
        inp >> h >> k >> l
                >> Imeas
                >> sigImeas
                >> Fcalc
                >> phicalc;
        if (inp.fail()) break;
        HKL hkl (h,k,l);
        FCFcomponents data = {hkl, Imeas, sigImeas, Fcalc, phicalc};
        fcfdata_.push_back(data);
    }
}

void FCFfile::readsymop(std::ifstream& inp) {
    Int3x3 R;
    const unsigned int linewidth (81);
    char line[linewidth];
    unsigned int firsttick, lasttick;
    unsigned int firstcomma, lastcomma;
    
    inp.getline(line, linewidth);
    std::string symop (line);
    std::string entry;
    
    /* sanity checks */
#ifdef __SUNPRO_CC
    int n1 (0), n2(0);
    std::count(symop.begin(), symop.end(), '\'', n1);
    std::count(symop.begin(), symop.end(), ',', n2);
    if (n1 != 2 || n2 != 2) {
        throw myExcepts::Format::Symop(symop.c_str());
    }
#else
    if (std::count(symop.begin(), symop.end(), '\'') != 2 || 
            std::count(symop.begin(), symop.end(), ',') != 2) {
        throw myExcepts::Format::Symop(symop.c_str());
    }
#endif
    firsttick = symop.find('\'');
    lasttick = symop.rfind('\'');
    firstcomma = symop.find(',');
    lastcomma = symop.rfind(',');
    
    std::string symopString = symop.substr(firsttick+1, lasttick - firsttick -1);
    symopsStrings_.push_back(symopString);

    if (symop.find("'x, y, z'") == std::string::npos) {
        /* extract first row data */
        entry = symop.substr(firsttick + 1, firstcomma - firsttick - 1);
        tosymop(entry, R, 0);
        /* extract second row data */
        entry = symop.substr(firstcomma + 1, lastcomma - firstcomma - 1);
        tosymop(entry, R, 1);
        /* extract third row data */
        entry = symop.substr(lastcomma + 1, lasttick - 1);
        tosymop(entry, R, 2);

        symops_.push_back(R);
    }
    try {
        readsymop(inp);
    }
    catch (myExcepts::Format::Symop& s) {
        return;
    }
}

void FCFfile::tosymop (const std::string& symstring, Int3x3& symop, int row) const{
    if (symstring.find("-x") != std::string::npos) {
        symop(row, 0) = -1;
    } else if (symstring.find("x") != std::string::npos) {
        symop(row, 0) = 1;
    }
    if (symstring.find("-y") != std::string::npos) {
        symop(row, 1) = -1;
    } else if (symstring.find("y") != std::string::npos) {
        symop(row, 1) = 1;
    }
    if (symstring.find("-z") != std::string::npos) {
        symop(row, 2) = -1;
    } else if (symstring.find("z") != std::string::npos) {
        symop(row, 2) = 1;
    }
    size_t dividx = symstring.find('/');
    
    if ( dividx != std::string::npos) {
        std::istringstream inp (symstring.substr(dividx-2,4));
        int nominator, denominator;
        char slash;
        inp >> nominator >> slash >> denominator;
        symop(row) = 1.0*nominator/denominator;
    }
}

std::vector<Reflex> FCFfile::reflections(int mapcoeff) const {
    std::vector<Reflex> reflections;
    
    std::vector<FCFcomponents>::const_iterator it;
    for (it=fcfdata_.begin(); it != fcfdata_.end(); ++it) {
                // modify F according to options
        /* sanity settings to avoid negative intensities */
        double Imeas = std::max(it->Imeas, 0.0);
        double Fmeas = std::sqrt(Imeas);
        double sigFmeas = 0.5*it->sigImeas / Fmeas;
        double F;

        if (mapcoeff == 0) {
            F = Fmeas - it->Fcalc;
        }
        else {
            F = mapcoeff*Fmeas - (mapcoeff-1)*it->Fcalc;
        }
        if (weightWeaks_ >= 0.0) {
            double scale = sigFmeas/it->Fcalc;
            scale = 1.0 /
                (1.0 + weightWeaks_ * std::pow(scale, 4));
            F *= scale;
        }
        
        Reflex reflex (it->hkl, F, it->phicalc);
        reflections.push_back(reflex);
    }
    return reflections;

}

std::vector<Reflex> FCFfile::read_fadata(const char* filename) const {
    int h,k,l;
    double dF, sigdF, phi;
    std::vector<Reflex> fareflections;
    
    std::ifstream inp (filename);
    if (! inp.is_open()) {
        std::cout << "*** Error: Unable to open file \'"
                << filename << "\'\n";
        throw myExcepts::IO::Filename (filename);
    }
    
    unsigned int linenumber(0);
    while (true) {
        inp >> h >> k >> l >> dF >> sigdF >> phi;
        ++linenumber;
        if ( h == 0 && k == 0 && l == 0)        break;
        if (inp.eof()) break;
        if (inp.fail()) {
            std::cout << "*** Error: Cannot interpret _fa-file in line "
                    << linenumber << ". Make sure each\n"
                    << "           line consists of h k l deltaF sigDelteF phi\n";
            throw myExcepts::Format::FA (filename);
        }
        HKL hkl (h,k,l);
        Reflex r(hkl, dF, phi);
        fareflections.push_back(r);
    }
    std::cout << " ----> Read " << fareflections.size() <<
            " reflections from file \'" << filename << "\'\n";
    return fareflections;
}